import shutil
import sys
from pathlib import Path
from typing import Dict, List, Optional

import tomli
from pydantic import BaseModel

from acac.util import UTF_8, confirm_yN, console

ACAC_TOML = Path("acac.toml")
DEFAULT_ACAC_TOML = Path(__file__).parent / "default_acac.toml"


class Create(BaseModel):
    post_create_commands: List[str] = []
    clipboard_message = ""


class Judge(BaseModel):
    copy_source_code_when_ac = False
    clipboard_message = ""


class LanguageSetting(BaseModel):
    class Commands(BaseModel):
        pre_execute: List[str] = []
        execute: str
        post_execute: List[str] = []

    source_file_name: str
    template_file_path: Optional[Path]
    commands: Commands


class Language(BaseModel):
    default: str
    settings: Dict[str, LanguageSetting]


class Config(BaseModel):
    create: Create
    judge: Judge
    language: Language


def init() -> None:
    if ACAC_TOML.exists():
        console.print("acac.toml がこのディレクトリ内に既に存在します。")
        return
    shutil.copy(DEFAULT_ACAC_TOML, ACAC_TOML)
    console.print("[bold]Created:", ACAC_TOML)


def load() -> Config:
    if not ACAC_TOML.exists():
        console.print("acac.toml がこのディレクトリ内に見つかりませんでした。")
        if confirm_yN("acac.toml を作成しますか？"):
            init()
        else:
            sys.exit("Bye.")
    return Config(**tomli.loads(ACAC_TOML.read_text(encoding=UTF_8)))



from __future__ import annotations

import os
from pathlib import Path

from typing_extensions import Literal

from acac import algo_method, atcoder, create, judge
from acac.config import Config
from acac.share import Folder, ProblemType


def main(args: list[str], config: Config) -> None:
    url = args[0]
    problem_type = get_problem_type(url)
    lang_name = get_lang_name(args, config.language.default)
    source_file_name = get_source_file_name(
        args, config.language.settings[lang_name].source_file_name
    )
    dir_path = Path(get_dir_path(url))

    folder = Folder(
        dir_path=dir_path,
        in_=dir_path / "in",
        out=dir_path / "out",
        cache_html=dir_path / "cache.html",
        metadata_toml=dir_path / "metadata.toml",
        source_file=dir_path / source_file_name,
    )

    replace_map = {
        "${dir_path}": str(dir_path),
        "${lang}": lang_name,
        "${source_file_name}": source_file_name,
        "${url}": url,
    }

    mode = get_mode(args)

    if mode == "create":
        create.main(url, folder, problem_type, lang_name, config, replace_map)
    elif mode == "manual":
        create.main(
            url, folder, problem_type, lang_name, config, replace_map, manual=True
        )
    elif mode == "judge":
        judge.main(url, folder, problem_type, lang_name, config, replace_map)


def get_problem_type(url: str) -> ProblemType:
    if url.startswith(algo_method.BASE_URL):
        return "algo_method"
    if url.startswith(atcoder.BASE_URL):
        return "atcoder"
    return "else"


def get_after_equal_option(
    args: list[str], prefix: tuple[str, ...], default: str
) -> str:
    return next((x.split("=")[1] for x in args[::-1] if x.startswith(prefix)), default)


def get_lang_name(args: list[str], default: str) -> str:
    return get_after_equal_option(args, ("-l=", "--lang=", "lang="), default)


def get_source_file_name(args: list[str], default: str) -> str:
    return get_after_equal_option(args, ("-s=", "--source=", "source="), default)


def get_dir_path(url: str) -> str:
    if url.startswith(("http://", "https://")):
        return os.path.join(*url.split("/")[2:])
    else:
        return os.path.join(*url.split("/"))


def get_mode(args: list[str]) -> Literal["create", "judge", "manual"]:
    for x in args[::-1]:
        if x in {"-c", "--create"}:
            return "create"
        elif x in {"-j", "--judge"}:
            return "judge"
        elif x in {"-m", "--manual"}:
            return "manual"
    return "create"



from __future__ import annotations

import os
from pathlib import Path

from pydantic import BaseModel
from typing_extensions import Literal

from acac.util import replaced

IorO = Literal["入", "出"]
ProblemType = Literal["algo_method", "atcoder", "else"]


class Folder(BaseModel):
    dir_path: Path
    in_: Path
    out: Path
    cache_html: Path
    metadata_toml: Path
    source_file: Path


def expand_command(command: str, replace_map: dict[str, str]) -> list[str]:
    return [*map(os.path.expanduser, replaced(command, replace_map).split())]



from __future__ import annotations

from functools import partial
from typing import Callable

from bs4 import BeautifulSoup, Tag

from acac.share import IorO
from acac.util import get_next_tag_text, get_tags

BASE_URL = "https://atcoder.jp/contests/"


def compose_get_samples(soup: BeautifulSoup) -> Callable[[IorO], list[str]]:
    def gs(
        soup: BeautifulSoup,
        tag_name: str,
        next_tag_name: str,
        func: Callable[[Tag, str], str],
        i_or_o: IorO,
    ) -> list[str]:
        return [
            func(tag, next_tag_name)
            for tag in get_tags(soup, tag_name)
            if f"{i_or_o}力例" in tag.text
            # if f"Sample {'Input' if i_or_o=='入' else 'Output'}" in tag.text
        ]

    return partial(gs, soup, "h3", "pre", get_next_tag_text)


def get_ac_url(url: str, lang: str) -> str:
    language_name = {
        "bash": "Bash",
        "c": "C",
        "cpp": "C%2B%2B",
        "cs": "C%23",
        "csharp": "C%23",
        "go": "Go",
        "golang": "Go",
        "haskell": "Haskell",
        "hs": "Haskell",
        "java": "Java",
        "javascript": "JavaScript",
        "js": "JavaScript",
        "php": "PHP",
        "py": "Python3",
        "pypy": "Python3",
        "pypy2": "Python2",
        "pypy3": "Python3",
        "python": "Python3",
        "python2": "Python2",
        "python3": "Python3",
        "rb": "Ruby",
        "rs": "Rust",
        "ruby": "Ruby",
        "rust": "Rust",
        "ts": "TypeScript",
        "typescript": "TypeScript",
        "zsh": "Zsh",
    }.get(lang, "")
    return f"{BASE_URL}{get_contest_name(url)}/submissions?f.Task={get_task_name(url)}&f.LanguageName={language_name}&f.Status=AC&f.User="


def get_contest_name(url: str) -> str:
    return url.split("/")[4]


def get_task_name(url: str) -> str:
    return url.split("/")[-1]



from __future__ import annotations

import platform
from pathlib import Path

from acac import __version__, config, core
from acac.util import console, includes

HELP_MESSAGE = """\
競プロ便利ツール。

[bold]Usage:[/bold]
  acac <url> \\[options]
  acac init                              Create configuration file "acac.toml".

[bold]Options:[/bold]
  [blue]-c, --create[/blue]                           Create environment mode. (default)
  [blue]-j, --judge[/blue]                            Judge mode.
  [blue]-m, --manual[/blue]                           URL にアクセスせず、HTML ファイルを手動で配置してテストケースを作成するモード。

  [blue]-l, --lang,   lang=LANG_NAME[/blue]           言語を指定する。
  [blue]-s, --source, source=SOURCE_FILE_NAME[/blue]  ソースファイル名を指定する。

[bold]Global options:[/bold]
  [blue]-h, --help[/blue]                             Show this help message and exit.
  [blue]-V, --version[/blue]                          Show program's version number and exit.

See https://github.com/seijinrosen/acac for more information.\
"""


def print_help_message() -> None:
    console.print(HELP_MESSAGE)


def print_version() -> None:
    console.print("acac:  ", __version__)
    console.print("Python:", platform.python_version())
    console.print("from:  ", Path(__file__).parent)


def main(args: list[str]) -> None:
    if not args or includes(args, {"-h", "--help"}):
        print_help_message()
        return

    if includes(args, {"-V", "--version"}):
        print_version()
        return

    if args[0] == "init":
        config.init()
        return

    core.main(args, config.load())



from __future__ import annotations

import json
from itertools import takewhile

from bs4 import BeautifulSoup

from acac.share import IorO

BASE_URL = "https://algo-method.com/tasks/"


def get_samples(soup: BeautifulSoup, i_or_o: IorO) -> list[str]:
    it = iter(extract_script_body(soup).splitlines())
    ret: list[str] = []
    for row in it:
        if not row.startswith(f"#### {i_or_o}力例"):
            continue
        next(it)
        sample_texts = takewhile(lambda x: not x.startswith("```"), it)
        ret.append("\n".join(sample_texts))
    return ret


def extract_script_body(soup: BeautifulSoup) -> str:
    script_tag = soup.find("script", {"id": "__NEXT_DATA__"})
    script_text = script_tag.text if script_tag else ""
    script_json = json.loads(script_text)
    return script_json["props"]["pageProps"]["tasks"]["body"]


def get_ac_url(url: str, lang: str) -> str:
    task_id = url.split("/")[-1]
    language = {
        "c": 5,
        "cpp": 1,
        "cs": 7,
        "csharp": 7,
        "go": 9,
        "golang": 9,
        "haskell": 15,
        "hs": 15,
        "java": 6,
        "javascript": 8,
        "js": 8,
        "php": 17,
        "py": 2,
        "pypy": 16,
        "pypy3": 16,
        "python": 2,
        "python3": 2,
        "rb": 3,
        "rs": 13,
        "ruby": 3,
        "ruby3": 3,
        "rust": 13,
        "ts": 11,
        "typescript": 11,
    }.get(lang)
    if language is None:
        return f"{url}/submissions?id={task_id}&status=AC&page=0"
    return f"{url}/submissions?id={task_id}&language={language}&status=AC&page=0"



from __future__ import annotations

import shutil
from pathlib import Path

from pydantic import BaseModel

from acac import algo_method, atcoder
from acac.config import Config
from acac.share import Folder, ProblemType, expand_command
from acac.util import (
    UTF_8,
    confirm_yN,
    console,
    copy2clip_with_log,
    dump_to_toml,
    get_soup,
    get_title,
    replaced,
    request_bytes,
    run_with_log,
)


class Metadata(BaseModel):
    title: str
    url: str


def main(
    url: str,
    folder: Folder,
    problem_type: ProblemType,
    lang_name: str,
    config: Config,
    replace_map: dict[str, str],
    manual: bool = False,
) -> None:
    if not folder.dir_path.exists():
        folder.dir_path.mkdir(parents=True)
        console.print("[bold]mkdir:", folder.dir_path)

    if not folder.source_file.exists():
        template_file = config.language.settings[lang_name].template_file_path
        if template_file and template_file.expanduser().exists():
            shutil.copy(template_file.expanduser(), folder.source_file)
            console.print("[bold]Copied:", template_file, "->", folder.source_file)
        else:
            folder.source_file.touch()
            console.print("[bold]touch:", folder.source_file)

    if not folder.cache_html.exists():
        if manual:
            console.print("マニュアルモード")
            if confirm_yN("HTML ファイルを配置しましたか？"):
                html_file = next(folder.dir_path.glob("*.html"))
                html_file.rename(folder.cache_html)
                console.print("[bold]Renamed:", html_file, "->", folder.cache_html)
        elif problem_type == "atcoder":
            folder.cache_html.write_bytes(request_bytes(url + "?lang=ja"))
            console.print("[bold]Dumped:", folder.cache_html)
        else:
            folder.cache_html.write_bytes(request_bytes(url))
            console.print("[bold]Dumped:", folder.cache_html)

    soup = get_soup(folder.cache_html.read_bytes())

    if not folder.metadata_toml.exists():
        dump_to_toml(Metadata(title=get_title(soup), url=url), folder.metadata_toml)
        console.print("[bold]Created:", folder.metadata_toml)

    if problem_type == "algo_method":
        i_samples = algo_method.get_samples(soup, "入")
        o_samples = algo_method.get_samples(soup, "出")
        dump_samples(i_samples, folder.in_)
        dump_samples(o_samples, folder.out)
    else:
        get_samples = atcoder.compose_get_samples(soup)
        dump_samples(get_samples("入"), folder.in_)
        dump_samples(get_samples("出"), folder.out)

    for cmd in config.create.post_create_commands:
        run_with_log(expand_command(cmd, replace_map), check=True)

    if config.create.clipboard_message:
        copy2clip_with_log(replaced(config.create.clipboard_message, replace_map))


def dump_samples(samples: list[str], io_dir: Path) -> None:
    if not io_dir.exists():
        io_dir.mkdir()
        console.print("[bold]mkdir:", io_dir)

    for i, sample_str in enumerate(samples, start=1):
        file = io_dir / f"{i:02}.txt"
        if not file.exists():
            if sample_str == "":
                file.touch()
                console.print("[bold]touch:", file)
            else:
                file.write_text(sample_str + "\n", encoding=UTF_8)
                console.print("[bold]Created:", file)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['acac']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4>=4.11.1,<5.0.0',
 'lxml>=4.9.1,<5.0.0',
 'pydantic>=1.10.1,<2.0.0',
 'pyperclip>=1.8.2,<2.0.0',
 'readchar>=4.0.3,<5.0.0',
 'requests>=2.28.1,<3.0.0',
 'rich>=12.5.1,<14.0.0',
 'tomli-w>=1.0.0,<2.0.0',
 'tomli>=2.0.1,<3.0.0',
 'typing-extensions>=4.3.0,<5.0.0']

entry_points = \
{'console_scripts': ['acac = acac.__main__:main']}

setup_kwargs = {
    'name': 'acac',
    'version': '0.4.0',
    'description': '競プロ便利 CLI ツール。AtCoder と アルゴ式 に対応。',
    'long_description': '# acac\n\n競プロ便利 CLI ツール。[AtCoder](https://atcoder.jp/) と [アルゴ式](https://algo-method.com/) に対応。\n\n[![PyPI](https://img.shields.io/pypi/v/acac)](https://pypi.python.org/pypi/acac)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/acac)](https://pypi.python.org/pypi/acac)\n[![Tests](https://github.com/seijinrosen/acac/actions/workflows/tests.yml/badge.svg)](https://github.com/seijinrosen/acac/actions/workflows/tests.yml)\n[![CodeQL](https://github.com/seijinrosen/acac/actions/workflows/codeql.yml/badge.svg)](https://github.com/seijinrosen/acac/actions/workflows/codeql.yml)\n[![codecov](https://codecov.io/gh/seijinrosen/acac/branch/main/graph/badge.svg)](https://codecov.io/gh/seijinrosen/acac)\n[![Downloads](https://pepy.tech/badge/acac)](https://pepy.tech/project/acac)\n[![Downloads](https://pepy.tech/badge/acac/month)](https://pepy.tech/project/acac)\n[![Downloads](https://pepy.tech/badge/acac/week)](https://pepy.tech/project/acac)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n＊現在 Pre-release のため、挙動やコマンドは変更される場合があります。\n\n## 概要\n\n競技プログラミングの過去問を解くときの（個人的に）典型的なワークフローを CLI として自動化したものです。\n\n過去問だけでなく開催中のコンテストでも使えますが、ログイン機能は実装されていないため、手動で HTML ファイルを取得する必要があります。\n\n## インストール\n\nPython 3.7 以上がインストールされていれば利用可能です。\n\n```sh\npip install acac\n```\n\n## 事前準備\n\n作業ディレクトリに移動して、`acac init` を実行します。\n\n```sh\n# 例\nmkdir kyopro\ncd kyopro\nacac init\n```\n\n`acac.toml` が作成されます。これが設定ファイルです。\n\n## 使用例\n\n1. まず、ブラウザで問題ページ（例えば、[ABC 280 A - Pawn on a Grid](https://atcoder.jp/contests/abc280/tasks/abc280_a)）にアクセスします。\n\n1. URL をコピーします\n\n   - 使用可能な場合、以下のショートカットキーが便利です。\n     - Windows: <kbd>Ctrl</kbd>+<kbd>L</kbd>, <kbd>Ctrl</kbd>+<kbd>C</kbd>\n     - Mac: <kbd>command</kbd>+<kbd>L</kbd>, <kbd>command</kbd>+<kbd>C</kbd>\n\n1. ターミナルで以下のようなコマンドを実行すると、問題用のフォルダ（以下、問題フォルダ）に環境が自動作成されます。\n\n   ```sh\n   acac https://atcoder.jp/contests/abc280/tasks/abc280_a\n   ```\n\n   <details><summary>処理の詳細</summary>\n\n   - 問題フォルダを作成します。\n   - ソースコードのテンプレートファイルが用意されていれば、そのファイルをコピーします。そうでなければ、ソースコード用の空ファイルを作成します。\n   - （`cache.html` が無ければ）問題ページにアクセスし、HTML ファイルを `cache.html` として保存します。\n   - `metadata.toml` を作成します。\n     - 問題ページのタイトルと URL が格納されます。\n   - 問題ページ中からテストケースのサンプルを抽出し、テキストファイルとして保存します。\n   - `acac.toml` で設定したコマンドを実行します。\n   - `acac.toml` で設定したメッセージをクリップボードにコピーします。\n     - 私は Git のコミットメッセージを設定しています。\n\n   </details>\n\n1. コードを書いて問題を解きます。\n\n1. ターミナルで以下のようなコマンドを実行します。\n\n   ```sh\n   acac https://atcoder.jp/contests/abc280/tasks/abc280_a -j\n   ```\n\n   すると、以下のように処理されます。\n\n   - `acac.toml` で設定したコマンドを実行します（バージョン確認、コンパイル等）。\n   - 用意されたテストケースに対してジャッジを行います。\n   - `acac.toml` で設定したコマンドを実行します（クリーンアップ等）。\n   - すべて AC であれば、ソースコードがクリップボードにコピーされますので、ブラウザに貼り付けて提出してください。\n   - 「他の人の提出を確認しますか？」と聞かれるので、`y` と答えれば、同じ言語で AC した提出の一覧ページをブラウザで開きます。\n   - `acac.toml` で設定したメッセージをクリップボードにコピーします。\n\n## 設定ファイル\n\n私が実際に使用している設定ファイルは [こちら](https://github.com/seijinrosen/kyopro/blob/main/acac.toml) です。\n\n```toml\n# 設定ファイルの例\n\n[create]\n# 環境作成後に実行されるコマンドのリスト（以下は git add をして、VSCode でソースコード用のファイルを開いている）\npost_create_commands = [\n    "git add ${dir_path}/in ${dir_path}/out ${dir_path}/metadata.toml",\n    "code . ${dir_path}/${source_file_name}",\n]\n# 環境作成後にクリップボードにコピーされるメッセージ\nclipboard_message = "Create: ${url}"\n\n\n[judge]\n# ジャッジ後にソースコードをクリップボードにコピーするかどうか\ncopy_source_code_when_ac = true\n# ジャッジ後にクリップボードにコピーされるメッセージ\nclipboard_message = "AC: ${url} ${source_file_name}"\n\n\n[language]\n# デフォルトの使用言語\ndefault = "cpp"\n\n\n[language.settings.cpp]\n# ソースコードのファイル名\nsource_file_name = "main.cpp"\n# テンプレートファイルのパス\ntemplate_file_path = "templates/main.cpp"\n[language.settings.cpp.commands]\n# ジャッジ前に実行するコマンドのリスト（以下はバージョンを表示し、コンパイルしている）\npre_execute = [\n    "g++ --version",\n    "g++ ${dir_path}/${source_file_name} -o ${dir_path}/a.out",\n]\n# 実行コマンド\nexecute = "${dir_path}/a.out"\n# ジャッジ後に実行するコマンドのリスト（以下は `a.out` を削除している）\npost_execute = ["rm ${dir_path}/a.out"]\n\n\n[language.settings.python3]\n# ...\n```\n\n### `${var}` の置換リスト\n\n| 置換前              | 置換後                                         |\n| ------------------- | ---------------------------------------------- |\n| ${dir_path}         | 問題フォルダのパス                             |\n| ${lang}             | 言語名                                         |\n| ${source_file_name} | ソースコードのファイル名（パスではありません） |\n| ${url}              | 問題ページの URL                               |\n\n## コマンドオプション\n\n### モード指定\n\n| オプション   | モード                                                                  |\n| ------------ | ----------------------------------------------------------------------- |\n| -c, --create | 作業環境構築（デフォルト）                                              |\n| -j, --judge  | ジャッジ                                                                |\n| -m, --manual | URL にアクセスせず、HTML ファイルを手動で配置してテストケースを作成する |\n\nログインが必要な場合、`acac <url> -m` を実行後、問題フォルダに問題ページの HTML ファイルを配置してください。\n\n### その他\n\n`acac.toml` に指定したデフォルト値を一時的に上書きするような動きをします。イコールは必須です。\n\n| オプション                            | 上書きされるもの         |\n| ------------------------------------- | ------------------------ |\n| -l, --lang, lang=LANG_NAME            | 使用言語                 |\n| -s, --source, source=SOURCE_FILE_NAME | ソースコードのファイル名 |\n\n```sh\n# 例\nacac https://atcoder.jp/contests/abc280/tasks/abc280_a -l=python3 --source=main2.py\nacac https://atcoder.jp/contests/abc280/tasks/abc280_a -s=main2.py lang=python3 --judge\n```\n\n## コンセプト\n\n### なぜ `acac create <url>` や `acac judge <url>` のような一般的な CLI の慣例に沿っていないのか\n\n1. `acac <url>` で環境作成\n1. コードを書く\n1. ターミナルで <kbd>Ctrl</kbd>+<kbd>P</kbd>\n1. 末尾に `-j` をつけてジャッジ\n\nという流れを高速で行うためです。基本的に、一つの問題に対し複数のコマンドを実行することが多いので、URL のあとにコマンドやオプションを指定する方式を採っています。\n\n### 問題フォルダ構成が URL そのままで冗長なのはなぜか\n\n開発当初は `AtCoder/ABC/280/A/` のようなフォルダ構成にしていましたが、過去のコンテストの URL 規則との整合性や、未来への拡張性、[ghq](https://github.com/x-motemen/ghq) のような厳密性を保持するため、現在のような形にしました。\n',
    'author': 'seijinrosen',
    'author_email': '86702775+seijinrosen@users.noreply.github.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/seijinrosen',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import json
import subprocess
import sys
from pathlib import Path

import pyperclip
import requests
import tomli_w
from bs4 import BeautifulSoup, ResultSet, Tag
from pydantic import BaseModel
from readchar import key, readchar
from rich.console import Console
from rich.markup import escape

UTF_8 = "utf-8"
console = Console()


def includes(args: list[str], flags: set[str]) -> bool:
    return any(s in args for s in flags)


def replaced(s: str, d: dict[str, str]) -> str:
    for old, new in d.items():
        s = s.replace(old, new)
    return s


# pyperclip
def copy2clip_with_log(text: str) -> None:
    pyperclip.copy(text)  # type: ignore
    console.print("[bold]Copied to clipboard:", escape(text))


# subprocess
def run_with_log(
    cmd_args: list[str] | list[Path] | list[str | Path],
    capture_output: bool = False,
    check: bool = False,
    input: str | None = None,
    text: bool = False,
) -> tuple[str, str]:
    console.print("[bold]Running:", *cmd_args)
    cp = subprocess.run(
        args=cmd_args,
        capture_output=capture_output,
        check=check,
        input=input,
        text=text,
    )
    return cp.stdout, cp.stderr


# tomli
def dump_to_toml(obj: BaseModel, toml_path: Path) -> None:
    toml_path.write_text(tomli_w.dumps(json.loads(obj.json())), encoding=UTF_8)


# requests
def request_bytes(url: str, with_log: bool = True) -> bytes:
    if with_log:
        console.print("[bold]Requesting...:", url)
    response = requests.get(url)
    response.raise_for_status()
    if with_log:
        console.print("Success!:thumbs_up:", style="bold green")
    response.encoding = response.apparent_encoding
    return response.content.replace(b"\r\n", b"\n")


# bs4
def get_soup(markup: str | bytes) -> BeautifulSoup:
    return BeautifulSoup(markup, "lxml")


def get_title(soup: BeautifulSoup) -> str:
    return soup.title.text.strip() if soup.title else ""


def get_tags(soup: BeautifulSoup, tag_name: str) -> ResultSet[Tag]:
    return soup(tag_name)


def get_next_tag_text(tag: Tag, next_tag_name: str) -> str:
    next_tag = tag.find_next(next_tag_name)
    return next_tag.text.strip() if next_tag else ""


# readchar
def confirm_yN(prompt: str) -> bool:
    console.print("[green]?", prompt, "(y/N): ", end="", style="bold")
    try:
        c = readchar()
        if c == key.CTRL_D:
            raise EOFError
    except (KeyboardInterrupt, EOFError):
        sys.exit("\nBye.")
    is_y = c == "y"
    console.print("Yes" if is_y else "No")
    return is_y



from __future__ import annotations

import subprocess
import time
import webbrowser
from pathlib import Path

import pyperclip
from pydantic import BaseModel
from rich.markup import escape
from rich.table import Table

from acac import algo_method, atcoder
from acac.config import Config
from acac.share import Folder, ProblemType, expand_command
from acac.util import (
    UTF_8,
    confirm_yN,
    console,
    copy2clip_with_log,
    replaced,
    run_with_log,
)


class IOSample(BaseModel):
    in_file: Path
    in_text: str
    out_text: str


class Result(BaseModel):
    name: str
    input: str
    expected: str
    actual: str
    error: str
    is_accepted: bool
    execution_ms: int


def main(
    url: str,
    folder: Folder,
    problem_type: ProblemType,
    lang_name: str,
    config: Config,
    replace_map: dict[str, str],
) -> None:
    lang_setting = config.language.settings[lang_name]

    for cmd in lang_setting.commands.pre_execute:
        run_with_log(expand_command(cmd, replace_map), check=True)

    results = get_results(
        expand_command(lang_setting.commands.execute, replace_map),
        load_io_samples(folder.in_, folder.out),
    )

    for cmd in lang_setting.commands.post_execute:
        run_with_log(expand_command(cmd, replace_map), check=True)

    console.print(create_table(results))

    if all(r.is_accepted for r in results):
        console.print("All Completed! AC!!!:thumbs_up:", style="green")

        if config.judge.copy_source_code_when_ac:
            source_code = folder.source_file.read_text(encoding=UTF_8)
            pyperclip.copy(source_code)  # type: ignore
            console.print("[bold]Copied source code to clipboard:", folder.source_file)

        if problem_type in {"algo_method", "atcoder"} and confirm_yN("他の人の提出を確認しますか？"):
            ac_url = get_ac_url(problem_type, url, lang_name)
            webbrowser.open(ac_url)
            console.print("[bold]Opened:", ac_url)

        if config.judge.clipboard_message:
            copy2clip_with_log(replaced(config.judge.clipboard_message, replace_map))
    else:
        console.print("WA...:", end=" ", style="red")
        console.print(*[r.name for r in results if not r.is_accepted])


def load_io_samples(i_dir: Path, o_dir: Path) -> list[IOSample]:
    return [
        IOSample(
            in_file=i_file,
            in_text=i_file.read_text(encoding=UTF_8),
            out_text=o_file.read_text(encoding=UTF_8),
        )
        for i_file, o_file in zip(sorted(i_dir.iterdir()), sorted(o_dir.iterdir()))
    ]


def get_results(execute_command: list[str], io_samples: list[IOSample]) -> list[Result]:
    def get_result(io_sample: IOSample) -> Result:
        start = time.time()
        console.print("[bold]Running:", *execute_command, "<", io_sample.in_file)
        cp = subprocess.run(
            execute_command, capture_output=True, input=io_sample.in_text, text=True
        )
        return Result(
            name=io_sample.in_file.name,
            input=io_sample.in_text.rstrip(),
            expected=io_sample.out_text.rstrip(),
            actual=cp.stdout.rstrip(),
            error=cp.stderr,
            is_accepted=io_sample.out_text.strip() == cp.stdout.strip(),
            execution_ms=int((time.time() - start) * 1000),
        )

    return [get_result(x) for x in io_samples]


def create_table(results: list[Result]) -> Table:
    table = Table(
        "name", "input", "expected", "actual", "error", header_style="bold magenta"
    )
    table.add_column("result", style="bold", justify="center")
    table.add_column("time(ms)", justify="right")
    for r in results:
        table.add_row(
            r.name,
            r.input,
            r.expected,
            r.actual,
            escape(r.error),
            "[green]AC" if r.is_accepted else "[red]WA",
            str(r.execution_ms),
        )
    return table


def get_ac_url(problem_type: ProblemType, url: str, lang: str) -> str:
    if problem_type == "algo_method":
        return algo_method.get_ac_url(url, lang)
    if problem_type == "atcoder":
        return atcoder.get_ac_url(url, lang)
    return ""
